/**
@file
@date        23/06/2014
@author      Josef Hofer
@copyright   CERN (EN-ICE-SCD)

@brief System Integrity plugin for the unLHCLoggingDB component.

Checks whether data transfer between RDB-Archive and LoggingDB is running properly.

Usage: Public

PVSS manager usage: CTRL
*/


//--------------------------------------------------------------------------------------------------------------------------------
// constant declaration
//------------------

// String constants for adressing datapoints
public const string c_SysInt_LHCLoggingDB = "LHCLoggingDB";
public const string LHCLoggingDB_pattern = "LHCLoggingDB";
public const string LDB_pattern = "LDB"; // shorter pattern for LHCLoggingDB

// Additional commands for function unSystemIntegrity_LHCLoggingDB_HandleCommand
public const int c_SysInt_LHCLoggingDB_CreateTestDps = 20;
public const int c_SysInt_LHCLoggingDB_RemoveTestDps = 21;

// Indexes of settings in .config.data
private const int c_SysInt_LHCLoggingDB_ConfigIndex_Count                 = 4;
private const int c_SysInt_LHCLoggingDB_ConfigIndex_LoopDelay             = 1;
private const int c_SysInt_LHCLoggingDB_ConfigIndex_MaxAgeTransferJobs    = 2;
private const int c_SysInt_LHCLoggingDB_ConfigIndex_MaxAgeTransferHistory = 3;
private const int c_SysInt_LHCLoggingDB_ConfigIndex_HourFullConsistencyCheck = 4;

// Default values for .config.data
public const unsigned c_SysInt_LHCLoggingDB_DefaultConfig_LoopDelay             = 240;  // [s]
public const unsigned c_SysInt_LHCLoggingDB_DefaultConfig_MaxAgeTransferJobs    = 600;  // [s]
public const unsigned c_SysInt_LHCLoggingDB_DefaultConfig_MaxAgeTransferHistory = 600;  // [s]
public const unsigned c_SysInt_LHCLoggingDB_DefaultConfig_HourFullConsistencyCheck = 8; // [h]

// Default values for hard coded settings

// Max. attempts to read ELEMENTS table before registering a signal
private const unsigned c_SysInt_LHCLoggingDB_DefaultConfig_RegistrationAttempts  = 5;    // [#]
// Wait time after unsucessfull attempt to read ELEMENTS table before doing next attempt
private const unsigned c_SysInt_LHCLoggingDB_DefaultConfig_RegistrationWaitTime  = 3;    // [s]

// Message indexes in LHCLoggingDB catalogue
private const int c_SysInt_LHCLoggingDB_Catalogue_ErrorNotFound             = 1;
private const int c_SysInt_LHCLoggingDB_Catalogue_ErrorLoadDbSettings       = 2;
private const int c_SysInt_LHCLoggingDB_Catalogue_ErrorOpenDbConnection     = 3;
private const int c_SysInt_LHCLoggingDB_Catalogue_ErrorDbSchemaNotFound     = 4;
private const int c_SysInt_LHCLoggingDB_Catalogue_ErrorCreatingTestDps      = 5;
private const int c_SysInt_LHCLoggingDB_Catalogue_ErrorDeletingTestDps      = 6;
private const int c_SysInt_LHCLoggingDB_Catalogue_ErrorRegisteringTestDps   = 7;
private const int c_SysInt_LHCLoggingDB_Catalogue_ErrorUnregisteringTestDps = 8;
private const int c_SysInt_LHCLoggingDB_Catalogue_ConfigDataInconsistent    = 9;
private const int c_SysInt_LHCLoggingDB_Catalogue_ConfigDataHasBeenReset    = 10;
private const int c_SysInt_LHCLoggingDB_Catalogue_ErrorStoppingThread       = 11;
private const int c_SysInt_LHCLoggingDB_Catalogue_ErrorFullConsistencyCheck = 12;

// Integrity classes
public const string c_SysInt_LHCLoggingDB_IntegrityClass_TransferJobs     = "LHCLoggingDB_TransferJobs";
public const string c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints   = "LHCLoggingDB_TestDatapoints";
public const string c_SysInt_LHCLoggingDB_IntegrityClass_DataConsistency  = "LHCLoggingDB_DataConsistency";
public const string c_SysInt_LHCLoggingDB_IntegrityClass_FullConsistency  = "LHCLoggingDB_FullConsistency";

// Alert levels of the Integrity classes
private const unsigned c_SysInt_LHCLoggingDB_TransferJobs_ErrorQuery            = 70;
private const unsigned c_SysInt_LHCLoggingDB_TransferJobs_ErrorTimeout          = 20;
private const unsigned c_SysInt_LHCLoggingDB_TransferJobs_ErrorDisabled         = 10;
private const unsigned c_SysInt_LHCLoggingDB_TransferJobs_OK                    =  0;

private const unsigned c_SysInt_LHCLoggingDB_TestDatapoints_ErrorQuery          = 70;
private const unsigned c_SysInt_LHCLoggingDB_TestDatapoints_ErrorRegistration   = 40;
private const unsigned c_SysInt_LHCLoggingDB_TestDatapoints_ErrorTransfer       = 30;
private const unsigned c_SysInt_LHCLoggingDB_TestDatapoints_OK                  =  0;

private const unsigned c_SysInt_LHCLoggingDB_DataConsistency_ErrorQuery         = 70;
private const unsigned c_SysInt_LHCLoggingDB_DataConsistency_ErrorMissingDpe    = 10;
private const unsigned c_SysInt_LHCLoggingDB_DataConsistency_ErrorArchiveGroup  =  5;
private const unsigned c_SysInt_LHCLoggingDB_DataConsistency_OK                 =  0;

private const unsigned c_SysInt_LHCLoggingDB_FullConsistency_ErrorDuringCheck   = 70;
private const unsigned c_SysInt_LHCLoggingDB_DataConsistency_ErrorIncosistent   = 10;
private const unsigned c_SysInt_LHCLoggingDB_FullConsistency_OK                 =  0;

// String constants
private const string c_SysInt_LHCLoggingDB_TestDatapoint_Description = "System integrity test datapoint for LHCLoggingDB";

//--------------------------------------------------------------------------------------------------------------------------------
// declaration of global variables
//------------------

private global dyn_string g_SysInt_LHCLoggingDB_ActiveIntegrityClasses;   // only needed for panel to show activated integrity classes
private global string g_SysInt_LHCLoggingDB_mySystemDbSchema;             // db schema of this system
private global int g_SysInt_LHCLoggingDB_monitoringThreadId;              // ID of the monitoring thread

private global unsigned g_SysInt_LHCLoggingDB_loopDelay;                  // [s] run checks according to this value
private global unsigned g_SysInt_LHCLoggingDB_maxAgeTransferJobs;         // [s] transfer jobs have to run within this time
private global unsigned g_SysInt_LHCLoggingDB_maxAgeTransferHistory;      // [s] transfer history has to be updated within this time

private global bool g_SysInt_LHCLoggingDB_fullConsistencyCheckLocked;     //
private global int g_SysInt_LHCLoggingDB_fullConsistencyCheckDaySecond;   // [s] start time of full consistency check for every day

private global bool g_SysInt_LHCLoggingDB_mutex;


//---------------------------------------------------------------------------------------------------------------------------------

/**
* Initialise system integrity plugin for LHCLoggingDB.
*
* @param dsResult   output, not used
*/
public void unSystemIntegrity_LHCLoggingDB_Initialize(dyn_string &dsResult)
{
    unSystemIntegrity_LHCLoggingDB_getConfigData(true);

    unSystemIntegrity_LHCLoggingDB_initDatabase();

    // Look for active integrity classes and write them into global list
    dyn_string dsSysIntAlarmDps = dpNames(c_unSystemAlarm_dpPattern + LHCLoggingDB_pattern + "*", c_unSystemAlarm_dpType);
    int iDsSysIntAlarmDpsLen = dynlen(dsSysIntAlarmDps);
    for (int i = 1; i <= iDsSysIntAlarmDpsLen; i++) {
        string sAlarmDpToCheck = dpSubStr(dsSysIntAlarmDps[i], DPSUB_DP);
        dynAppend(g_SysInt_LHCLoggingDB_ActiveIntegrityClasses, sAlarmDpToCheck);
    }
}

//---------------------------------------------------------------------------------------------------------------------------------

/// API function to trigger full consistency check on a distributed system
public void unSystemIntegrity_LHCLoggingDB_triggerFullConsistencyCheck(
        string sSystemName,
        const dyn_string &dsConfigurations,   // list of configurations to check
        bool bGenerateTempImportFiles,        // compare temp generated import files
        bool bSetAlarmDP,                     // trigger system integrity error
        bool bWriteResultToDP)                // save report in datapoint
{
  dyn_string dsParameters;
  dsParameters[1] = dynStringToString(dsConfigurations);
  dsParameters[2] = bGenerateTempImportFiles;
  dsParameters[3] = bSetAlarmDP;
  dsParameters[4] = bWriteResultToDP;

  dpSetWait(sSystemName+LHCLoggingDB_pattern+UN_SYSTEMINTEGRITY_EXTENSION+".interface.command", UN_SYSTEMINTEGRITY_EXTENDED_DIAGNOSTIC,
    		      sSystemName+LHCLoggingDB_pattern+UN_SYSTEMINTEGRITY_EXTENSION+".interface.parameters", dsParameters);
}

//---------------------------------------------------------------------------------------------------------------------------------

/**
* Handle systemIntegrity commands from configuration panel
*
* @param sDpe1      input, dpe
* @param command    input, the systemIntegrity command
* @param sDpe2      input, dpe
* @param parameters input, the list of parameters of the systemIntegrity command
*/
public void unSystemIntegrity_LHCLoggingDB_HandleCommand(string sDpe1, int command, string sDpe2, dyn_string parameters)
{
    dyn_string dsExInfo;

    switch (command) {
        case UN_SYSTEMINTEGRITY_ADD: { // add the _UnSystemAlarm DP
            int iParametersLen = dynlen(parameters);
            for (int i = 1; i <= iParametersLen; i++) {
                unSystemIntegrity_LHCLoggingDB_manageIntegrityClass(parameters[i], true, false, true, dsExInfo);

                if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints + "*", parameters[i])) {
                    unSystemIntegrity_LHCLoggingDB_addTestDPs();
                }
            }
            // ensure that thread is running
            unSystemIntegrity_LHCLoggingDB_manageMonitoringThread("start");
            break;
        }
        case UN_SYSTEMINTEGRITY_DELETE: { // delete the _UnSystemAlarm DP
            int iParametersLen = dynlen(parameters);
            for (int i = 1; i <= iParametersLen; i++) {
                unSystemIntegrity_LHCLoggingDB_manageIntegrityClass(parameters[i], false, true, false, dsExInfo);

                if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints + "*", parameters[i])) {
                    unSystemIntegrity_LHCLoggingDB_removeTestDPs();
                }
            }
            delay(1);
            // stop thread if necessary
            unSystemIntegrity_LHCLoggingDB_manageMonitoringThread("stop");
            break;
        }
        case UN_SYSTEMINTEGRITY_ENABLE: { // enable the _UnSystemAlarm DP
            int iParametersLen = dynlen(parameters);
            for (int i = 1; i <= iParametersLen; i++) {
                unSystemIntegrity_LHCLoggingDB_manageIntegrityClass(parameters[i], false, false, true, dsExInfo);

                if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints + "*", parameters[i])) {
                    unSystemIntegrity_LHCLoggingDB_addTestDPs();
                }
            }
            // ensure that thread is running
            unSystemIntegrity_LHCLoggingDB_manageMonitoringThread("start");
            break;
        }
        case UN_SYSTEMINTEGRITY_DISABLE: { // disable the _UnSystemAlarm DP
            int iParametersLen = dynlen(parameters);
            for (int i = 1; i <= iParametersLen; i++) {
                unSystemIntegrity_LHCLoggingDB_manageIntegrityClass(parameters[i], false, false, false, dsExInfo);
            }
            delay(1);
            // stop thread if necessary
            unSystemIntegrity_LHCLoggingDB_manageMonitoringThread("stop");
            break;
        }
        case UN_SYSTEMINTEGRITY_DIAGNOSTIC: { // give the list of _UnSystemAlarm DP and the state
            dpSetWait(c_SysInt_LHCLoggingDB + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.commandResult", command,
                      c_SysInt_LHCLoggingDB + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.result", g_SysInt_LHCLoggingDB_ActiveIntegrityClasses);
            break;
        }
        case UN_SYSTEMINTEGRITY_EXTENDED_DIAGNOSTIC: {
            dyn_string dsConfigurations = stringToDynString(parameters[1]);
            bool bGenerateTempImportFiles = parameters[2];
            bool bSetAlarmDP = parameters[3];
            bool bWriteResultToDP = parameters[4];

            unSystemIntegrity_LHCLoggingDB_doFullConsistencyCheck(dsConfigurations, bGenerateTempImportFiles, bSetAlarmDP, bWriteResultToDP, dsExInfo);
            string sAnswer = (dynlen(dsExInfo)>0 ? "error" : "ok");

            dpSetWait(c_SysInt_LHCLoggingDB + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.commandResult", command,
                      c_SysInt_LHCLoggingDB + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.result", makeDynString(sAnswer));
            break;
        }
        case c_SysInt_LHCLoggingDB_CreateTestDps: {
            if (dpExists(c_unSystemAlarm_dpPattern + c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints)) {
                DebugTN("creation of test dps was triggered");
                unSystemIntegrity_LHCLoggingDB_addTestDPs();
            }
            break;
        }
        case c_SysInt_LHCLoggingDB_RemoveTestDps: {
            DebugTN("destruction of test dps was triggered");
            unSystemIntegrity_LHCLoggingDB_removeTestDPs();
            break;
        }
        default: {
            break;
        }
    }

    if (dynlen(dsExInfo) > 0) {
        if (isFunctionDefined("unMessageText_sendException")) {
            unMessageText_sendException("*", "*", g_SystemIntegrity_sSystemName + "unSystemIntegrity_LHCLoggingDB_HandleCommand", "user", "*", dsExInfo);
        }
        // handle any error in case the send message failed
        if (dynlen(dsExInfo) > 0) {
            DebugTN(g_SystemIntegrity_sSystemName + "unSystemIntegrity_LHCLoggingDB_HandleCommand", dsExInfo);
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Callback function to get most recent configuration data
*
* @param sDPe
* @param dsResult
*/
public void unSystemIntegrity_LHCLoggingDB_DataCallback(string sDPe, dyn_string &dsResult)
{
    unSystemIntegrity_LHCLoggingDB_getConfigData(false);
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Create systemIntegrity datapoint for LHCLoggingDB
*/
public void unSystemIntegrity_LHCLoggingDB_postInstall()
{
    DebugTN("unSystemIntegrity_LHCLoggingDB_postInstall(): performing post installation steps");

    // Create system integrity datapoint for LHC Logging DB
    const string sSysIntDp = c_SysInt_LHCLoggingDB + UN_SYSTEMINTEGRITY_EXTENSION;
    if (!dpExists(sSysIntDp)) {
        dpCreate(sSysIntDp, UN_SYSTEMINTEGRITY_DPTYPE);

        dpSetWait(
            sSysIntDp + ".config.functions",
            makeDynString("unSystemIntegrity_LHCLoggingDB_Initialize"
                         ,"unSystemIntegrity_LHCLoggingDB_HandleCommand"
                         ,"unSystemIntegrity_LHCLoggingDB_DataCallback"),
            sSysIntDp + ".config.configurationPanel",
            "vision/systemIntegrity/systemIntegrity_LHCLoggingDB_configuration.pnl",
            sSysIntDp + ".config.operationPanel",
            "vision/systemIntegrity/systemIntegrity_LHCLoggingDB_operation.pnl"
        );
    }

    // Get settings and reset them if they are not correct
    unSystemIntegrity_LHCLoggingDB_getConfigData(true);

    // Delete system alarm datapoint for transfer history (this alarm is obsolete for versions >= 1.1.0)
    const string obsoleteSytemAlarmDP = "_unSystemAlarm_LHCLoggingDB_TransferHistory";
    if(dpExists(obsoleteSytemAlarmDP)) {
        dyn_string dsExInfo;
        unSystemIntegrity_LHCLoggingDB_manageIntegrityClass(obsoleteSytemAlarmDP, false, true, false, dsExInfo);
    }

    // Recreate alert config of LHC Logging Db system alarm dps to update it in case of change in structure
    dyn_string dsAlarmDps = dpNames(c_unSystemAlarm_dpPattern + LHCLoggingDB_pattern + "*", c_unSystemAlarm_dpType);
    dyn_errClass err = getLastError();
    if (dynlen(err) == 0) {
        int iDsAlarmDpsLen = dynlen(dsAlarmDps);
        for (int i = 1; i <= iDsAlarmDpsLen; i++) {
            dyn_string dsExInfo;
            fwAlertConfig_delete(dsAlarmDps[i] + ".alarm", dsExInfo) ;
            unSystemIntegrity_LHCLoggingDB_addAlertConfig(dsAlarmDps[i] + ".alarm", dsExInfo);
        }
    }
    int retVal;
    DebugTN("Restarting UNICOS scripts");
    retVal = fwInstallationManager_command("RESTART", "WCCOActrl", "-f unicos_scripts.lst");
    if(retVal != 0) DebugTN("ERROR restarting UNICOS scripts");
}


//------------------------------------------------------------------------------------------------------------------------

/**
* Wrapper for the below function. Deprecated.
*/
private void unSystemIntegrity_LHCLoggingDB_initDabatase(bool printDebug = true)
{
    DebugTN("Function unSystemIntegrity_LHCLoggingDB_initDabatase is deprecated! Please use unSystemIntegrity_LHCLoggingDB_initDatabase instead.");
    unSystemIntegrity_LHCLoggingDB_initDatabase(printDebug);
}

/**
* Get database settings, db schema and check if db schema exists in database
*/
private void unSystemIntegrity_LHCLoggingDB_initDatabase(bool printDebug = true) synchronized(g_SysInt_LHCLoggingDB_mutex)
{
    string sDbName, sDbUser, sDbPass;

    // Get database settings
    if (unLHCLoggingDBSettings_getActiveDbConfig(sDbName, sDbUser, sDbPass) == false) {
        unSystemIntegrity_LHCLoggingDB_throwError("WARNING", c_SysInt_LHCLoggingDB_Catalogue_ErrorLoadDbSettings);
        return;
    }

    // Connect to database
    if (unLHCLoggingDBSql_openDbConnection(sDbName, sDbUser, sDbPass, printDebug) == false) {
        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorOpenDbConnection);
        return;
    }

    // Get db schemas of the connected db
    dyn_string dsExInfo;
    dyn_string dsDbSchemas = unLHCLoggingDBSql_getDbSchemas(dsExInfo);
    if (dynlen(dsExInfo) > 0) {
        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorDbSchemaNotFound);
        return;
    }

    // Get db schema of this system
    g_SysInt_LHCLoggingDB_mySystemDbSchema = unLHCLoggingDBSettings_getMySystemDbSchema();

    // Check if schema is known by the database
    if (dynContains(dsDbSchemas, g_SysInt_LHCLoggingDB_mySystemDbSchema) <= 0) {
        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorDbSchemaNotFound);
        return;
    }
}

//------------------------------------------------------------------------------------------------------------------------

private void unSystemIntegrity_LHCLoggingDB_addTestDPs() synchronized(g_SysInt_LHCLoggingDB_mutex)
{
    unSystemIntegrity_LHCLoggingDB_createTestDPs();
    unSystemIntegrity_LHCLoggingDB_registerTestDPs();
}

//------------------------------------------------------------------------------------------------------------------------

private void unSystemIntegrity_LHCLoggingDB_removeTestDPs() synchronized(g_SysInt_LHCLoggingDB_mutex)
{
    unSystemIntegrity_LHCLoggingDB_unregisterTestDPs();
    unSystemIntegrity_LHCLoggingDB_deleteTestDPs();
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Create a test datapoint for each data category, if it does not exist yet
*/
private void unSystemIntegrity_LHCLoggingDB_createTestDPs()
{
    // Get data categories for which we will create test datapoints
    dyn_string dsExInfo;
    dyn_dyn_mixed ddmDataCategories = unLHCLoggingDBSql_getDataCategoriesOfSchema(g_SysInt_LHCLoggingDB_mySystemDbSchema, dsExInfo);
    if (dynlen(dsExInfo) > 0) {
        unSystemIntegrity_LHCLoggingDB_throwError(
                "ERROR",
                c_SysInt_LHCLoggingDB_Catalogue_ErrorCreatingTestDps,
                "Data categories could not be read");
        return;
    }

    // Create test datapoints
    int len = dynlen(ddmDataCategories);
    for (int i = 1; i <= len; i++) {
        string sTestDP = c_SysInt_RDB_TestDp_Prefix + "_" + LDB_pattern + "_" + ddmDataCategories[i][1];
        string sTestDPold = c_SysInt_RDB_TestDp_Prefix + "_" + LHCLoggingDB_pattern + "_" + ddmDataCategories[i][1];

        if (!dpExists(sTestDP) && !dpExists(sTestDPold)) { // only create new test dp, if there is no existing one
            dpCreate(sTestDP, c_SysInt_RDB_TestDp_Type);
            if (!dpExists(sTestDP)) {
                unSystemIntegrity_LHCLoggingDB_throwError(
                        "ERROR",
                        c_SysInt_LHCLoggingDB_Catalogue_ErrorCreatingTestDps,
                        "Datapoint " + sTestDP + " not created");
                return;
            }
            dpSetDescription(sTestDP + c_SysInt_RDB_TestDpe_Value, c_SysInt_LHCLoggingDB_TestDatapoint_Description);
            unSystemIntegrity_LHCLoggingDB_addArchiveConfig(sTestDP + c_SysInt_RDB_TestDpe_Value);
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Register test dapaoints for logging
*/
private void unSystemIntegrity_LHCLoggingDB_registerTestDPs()
{
    dyn_string dsDpe, dsAlias, dsVarName, dsDescription, dsDataCategoryId, dsHierarchy;
    dyn_string dsElementId, dsUnit, dsRDBSource;
    dyn_anytype dsErrors;
    dyn_string dsExInfo;

    dyn_string dsTestDp = unSystemIntegrity_LHCLoggingDB_getTestDps();

    int len = dynlen(dsTestDp);
    for (int i = 1; i <= len; i++) {
        dynAppend(dsVarName,  unSystemIntegrity_LHCLoggingDB_getVarName(dsTestDp[i]));
        dynAppend(dsDpe, dpSubStr(dsTestDp[i] + c_SysInt_RDB_TestDpe_Value, DPSUB_SYS_DP_EL));
        dynAppend(dsDescription, c_SysInt_LHCLoggingDB_TestDatapoint_Description);
        dynAppend(dsHierarchy, "");
        dynAppend(dsAlias, "");
        string sDataCategory = unSystemIntegrity_LHCLoggingDB_getDataCategory(dsTestDp[i]);
        dynAppend(dsDataCategoryId, unLHCLoggingDBSql_getDataCategoryId(sDataCategory, dsExInfo));
    }

    if (dynlen(dsDpe) > 0) {
        unsigned attempts = 0;
        while (true) {
            if (unLHCLoggingDBParser_getElementData(dsDpe, dsDpe, dsAlias, dsElementId, dsUnit, dsRDBSource, dsErrors)) { break; }
            if (++attempts > c_SysInt_LHCLoggingDB_DefaultConfig_RegistrationAttempts) {
                unSystemIntegrity_LHCLoggingDB_throwError(
                        "ERROR",
                        c_SysInt_LHCLoggingDB_Catalogue_ErrorRegisteringTestDps,
                        "Error retrieving ELEMENT_ID");
                return;
            }
            delay(c_SysInt_LHCLoggingDB_DefaultConfig_RegistrationWaitTime);
        }

        string sUser, sPass;
        unLHCLoggingDBSql_getSystemIntegrityCredentials(sUser, sPass, dsExInfo);
        if (dynlen(dsExInfo) > 0) {
            unSystemIntegrity_LHCLoggingDB_throwError(
                    "ERROR",
                    c_SysInt_LHCLoggingDB_Catalogue_ErrorRegisteringTestDps,
                    "Error getting credentials");
            return;
        }

        if (false == unLHCLoggingDBSql_registerDataForLogging(dsElementId, dsVarName, dsRDBSource, dsUnit, dsDescription, dsDataCategoryId, dsHierarchy, sUser, sPass)) {
            unSystemIntegrity_LHCLoggingDB_throwError(
                    "ERROR",
                    c_SysInt_LHCLoggingDB_Catalogue_ErrorRegisteringTestDps,
                    "Error during registration");
            return;
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Delete test datapoints for each data category
*/
private void unSystemIntegrity_LHCLoggingDB_deleteTestDPs()
{
    dyn_string dsTestDP = unSystemIntegrity_LHCLoggingDB_getTestDps();

    int len = dynlen(dsTestDP);
    for (int i = 1; i <= len; i++) {
        dpDelete(dsTestDP[i]);
        dyn_errClass err = getLastError();
        if (dynlen(err) > 0) {
            unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorDeletingTestDps);
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Unregister test dapaoints from logging
*/
private void unSystemIntegrity_LHCLoggingDB_unregisterTestDPs()
{
    dyn_string dsVariableName = unSystemIntegrity_LHCLoggingDB_getTestDpVarNames();
    bool success = unLHCLoggingDBSql_unregisterVariablesFromLogging(dsVariableName);
    if(!success) {
        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorUnregisteringTestDps);
    }
}

//------------------------------------------------------------------------------------------------------------------------

/**
Add archive config for archive group EVENT to the test dpe

@param sTestDpe, input, name of the test datapoint
*/
private void unSystemIntegrity_LHCLoggingDB_addArchiveConfig(string sTestDpe) {
    // get archive class for "EVENT"
    string sArchiveClass = unSystemIntegrity_LHCLoggingDB_getArchiveClass("EVENT");
    if (sArchiveClass == "") { return; } // no archive class found

    // check if there is already an archive config
    int iArchiveType = 0;
    dpGet(sTestDpe + ":_archive.._type", iArchiveType);
    if (iArchiveType != 0) { return; } // archive config already exists

    // create archive config
    dyn_string dsExInfo;
    fwArchive_set(sTestDpe, sArchiveClass, DPATTR_ARCH_PROC_VALARCH, 0, 0, 0, dsExInfo);
    if (dynlen(dsExInfo)) {
        unSystemIntegrity_LHCLoggingDB_throwError(
                "ERROR",
                c_SysInt_LHCLoggingDB_Catalogue_ErrorCreatingTestDps,
                "Archive config not created for " + sTestDpe );
    }
}

//------------------------------------------------------------------------------------------------------------------------

/**
Get the name of the archive class which matches the pattern

@param sPattern, input, pattern of the archive class like 'EVENT'
@return name of the archive class
*/
private string unSystemIntegrity_LHCLoggingDB_getArchiveClass(string sPattern) {

    // Get all archive classes
    dyn_string dsArchiveClasses, dsArchiveGroupDps, dsExInfo;
    fwArchive_getAllRDBArchiveClasses(makeDynString(getSystemName()), dsArchiveClasses, dsArchiveGroupDps, dsExInfo);
    if (dynlen(dsExInfo) > 0) {
        unSystemIntegrity_LHCLoggingDB_throwError(
                "ERROR",
                c_SysInt_LHCLoggingDB_Catalogue_ErrorCreatingTestDps,
                "Archive groups could not be read");
        return "";
    }

    // Check if there is an archive class that matches the pattern
    string sArchiveClass = "";
    int iDsArchiveClassesLen = dynlen(dsArchiveClasses);
    for (int i = 1; i <= iDsArchiveClassesLen; i++) {
        if (patternMatch("*"+sPattern+"*", dsArchiveClasses[i])) {
            sArchiveClass = dsArchiveClasses[i];
            break;
        }
    }
    if (sArchiveClass == "") {
        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorCreatingTestDps, "Archive group 'EVENT' not found");
        return "";
    }

    return sArchiveClass;
}

//------------------------------------------------------------------------------------------------------------------------

/**
Get list of all variable names

@return list of test datapoints
*/
private dyn_string unSystemIntegrity_LHCLoggingDB_getTestDpVarNames() {
    dyn_string dsVariableName;
    dyn_string dsTestDp = unSystemIntegrity_LHCLoggingDB_getTestDps();

    // Calculate the variable name and append it to the result list
    int len = dynlen(dsTestDp);
    for (int i = 1; i <= len; i++) {
      string sVarName = unSystemIntegrity_LHCLoggingDB_getVarName(dsTestDp[i]);
      dynAppend(dsVariableName, sVarName);
    }

    return dsVariableName;
}

//------------------------------------------------------------------------------------------------------------------------

/**
Get list of all test datapoints

@return dsTestDp, list of test datapoints
*/
private dyn_string unSystemIntegrity_LHCLoggingDB_getTestDps() {
   dyn_string dsTestDp = dpNames(c_SysInt_RDB_TestDp_Prefix + "_" + LHCLoggingDB_pattern + "_" + "*", c_SysInt_RDB_TestDp_Type);
   dyn_string dsTestDp2= dpNames(c_SysInt_RDB_TestDp_Prefix + "_" + LDB_pattern + "_" + "*", c_SysInt_RDB_TestDp_Type);
   dynAppend(dsTestDp, dsTestDp2);
   return dsTestDp;
}

//------------------------------------------------------------------------------------------------------------------------

/**
Calculate the variable name from the dpe name of the test datapoint

@param sDpeName, input, name of the test dp
@return variable name
*/
private string unSystemIntegrity_LHCLoggingDB_getVarName(string sTestDpName) {
    // Remove system name
    sTestDpName = dpSubStr(sTestDpName,DPSUB_DP);
    dyn_string dsSplit = strsplit(sTestDpName,"_");

    // Extract 'LHCLoggingDB' or 'LDB' pattern
    string sPattern = dsSplit[2];

    // Extract the data category
    string sDataCategory = unSystemIntegrity_LHCLoggingDB_getDataCategory(sTestDpName);
    strreplace(sDataCategory, "/", "_"); // replace '/' by '_' because we don't use '/' for separation in DB

    // Return the variable name
    string sVarName = strtoupper(c_SysInt_RDB_TestDp_Prefix + "." + sPattern + "." + getSystemName() + sDataCategory);
    DebugFTN(1, __FUNCTION__+":"+__LINE__, "sVarName="+sVarName);
    return (sVarName);
}

//------------------------------------------------------------------------------------------------------------------------

/**
Extract data category string from test datapoint name

@param sDpeName, input, name of the test dp
@return data category
*/
private string unSystemIntegrity_LHCLoggingDB_getDataCategory(string sTestDpName) {
    // Remove system name
    sTestDpName = dpSubStr(sTestDpName,DPSUB_DP);
    dyn_string dsSplit = strsplit(sTestDpName,"_");

    // Extract 'LHCLoggingDB' or 'LDB' pattern
    string sPattern = dsSplit[2];

    // Return the data category
    return (substr(sTestDpName, strlen(c_SysInt_RDB_TestDp_Prefix + "_" + sPattern + "_")));
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Get configuration data and settings for this system integrity plugin
*
* @param   bResetOnError   input,  true: set default values on error, false: don't set default values on error
*/
private void unSystemIntegrity_LHCLoggingDB_getConfigData(bool bResetOnError)
{
    string sConfigDP = c_SysInt_LHCLoggingDB + UN_SYSTEMINTEGRITY_EXTENSION;
    dyn_string dsConfigData;

    if (dpExists(sConfigDP)) {
        dpGet(sConfigDP+".config.data", dsConfigData);

        if (dynlen(dsConfigData) == c_SysInt_LHCLoggingDB_ConfigIndex_Count) {
            g_SysInt_LHCLoggingDB_loopDelay             = dsConfigData[c_SysInt_LHCLoggingDB_ConfigIndex_LoopDelay];
            g_SysInt_LHCLoggingDB_maxAgeTransferJobs    = dsConfigData[c_SysInt_LHCLoggingDB_ConfigIndex_MaxAgeTransferJobs];
            g_SysInt_LHCLoggingDB_maxAgeTransferHistory = dsConfigData[c_SysInt_LHCLoggingDB_ConfigIndex_MaxAgeTransferHistory];
            unsigned temp = dsConfigData[c_SysInt_LHCLoggingDB_ConfigIndex_HourFullConsistencyCheck];
            g_SysInt_LHCLoggingDB_fullConsistencyCheckDaySecond = (temp*3600)%86400;
        } else {
            if (bResetOnError) {
                dpSetWait(sConfigDP+".config.data",
                          makeDynString(c_SysInt_LHCLoggingDB_DefaultConfig_LoopDelay
                                      , c_SysInt_LHCLoggingDB_DefaultConfig_MaxAgeTransferJobs
                                      , c_SysInt_LHCLoggingDB_DefaultConfig_MaxAgeTransferHistory
                                      , c_SysInt_LHCLoggingDB_DefaultConfig_HourFullConsistencyCheck)
                          );
                g_SysInt_LHCLoggingDB_loopDelay             = c_SysInt_LHCLoggingDB_DefaultConfig_LoopDelay;
                g_SysInt_LHCLoggingDB_maxAgeTransferJobs    = c_SysInt_LHCLoggingDB_DefaultConfig_MaxAgeTransferJobs;
                g_SysInt_LHCLoggingDB_maxAgeTransferHistory = c_SysInt_LHCLoggingDB_DefaultConfig_MaxAgeTransferHistory;
                g_SysInt_LHCLoggingDB_fullConsistencyCheckDaySecond = (c_SysInt_LHCLoggingDB_DefaultConfig_HourFullConsistencyCheck*3600)%86400;

                unSystemIntegrity_LHCLoggingDB_throwError("INFO", c_SysInt_LHCLoggingDB_Catalogue_ConfigDataHasBeenReset);
            } else {
                unSystemIntegrity_LHCLoggingDB_throwError("WARNING", c_SysInt_LHCLoggingDB_Catalogue_ConfigDataInconsistent);
            }
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

/**
* This function is used to create/destruct/enable/disable alarm datapoints
* It can also create/destruct the test datapoints
*
* @param sDp             input, name of the alert dp
* @param bCreate         input, create dp (true: yes; false: no)
* @param bDelete         input, delete dp (true: yes; false: no)
* @param bEnable         input, true: enable alarm; false: disable alarm
* @param dsExInfo   output, exception are returned here
*/
private void unSystemIntegrity_LHCLoggingDB_manageIntegrityClass(string sDp, bool bCreate, bool bDelete, bool bEnable, dyn_string &dsExInfo)
{
    // Remove the system name
    string sSysAlarmDp = unSystemIntegrity_removeSystemName(sDp);

    if (bCreate) {
        unSystemIntegrity_LHCLoggingDB_createAlertDp(sSysAlarmDp);
        unSystemIntegrity_LHCLoggingDB_addAlertConfig(sSysAlarmDp + ".alarm", dsExInfo);

        if (dynlen(dsExInfo) <= 0) {
            // add to active integrity classes
            _unSystemIntegrity_setList(g_SysInt_LHCLoggingDB_ActiveIntegrityClasses, sSysAlarmDp, true);
        }
    }

    // Do enabling and masking
    if (dynlen(dsExInfo) <= 0) {
        if (dpExists(sSysAlarmDp)) {
            dpSetWait(sSysAlarmDp + ".enabled", bEnable);
            unAlarmConfig_mask(sSysAlarmDp + ".alarm", !bEnable, dsExInfo);
        }
    }

    if (bDelete) {
        // remove dp from the alert list of applicationDP
        _unSystemIntegrity_modifyApplicationAlertList(sSysAlarmDp, false, dsExInfo);

        // remove alert class from list
        _unSystemIntegrity_setList(g_SysInt_LHCLoggingDB_ActiveIntegrityClasses, sSysAlarmDp, false);

        // delete alarm datapoint
        if (dpExists(sSysAlarmDp)) {
            dpDelete(sSysAlarmDp);
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Create system integrity alert datapoint for LHC Logging DB
*
* @param sSysAlarmDp    input, alarm dp to create
*/
private void unSystemIntegrity_LHCLoggingDB_createAlertDp(string sSysAlarmDp)
{
    if (!dpExists(sSysAlarmDp)) {
        dpCreate(sSysAlarmDp, c_unSystemAlarm_dpType);

        string sSysAlarmDpe = sSysAlarmDp + ".alarm";
        if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TransferJobs + "*", sSysAlarmDpe)) {
            dpSetDescription(sSysAlarmDpe, "LHCLoggingDB transfer jobs:");
        } else if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints + "*", sSysAlarmDpe)) {
            dpSetDescription(sSysAlarmDpe, "LHCLoggingDB test datapoints:");
        } else if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_DataConsistency + "*", sSysAlarmDpe)) {
            dpSetDescription(sSysAlarmDpe, "LHCLoggingDB data consistency:");
        } else if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_FullConsistency + "*", sSysAlarmDpe)) {
            dpSetDescription(sSysAlarmDpe, "LHCLoggingDB full consistency check:");
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Add discrete alert config to system alarm datapoint element
*
* @param sSysAlarmDpe    input, dpe for which we want to create the alert config
* @param dsExInfo   output, exception are returned here
*/
private void unSystemIntegrity_LHCLoggingDB_addAlertConfig(string sSysAlarmDpe, dyn_string &dsExInfo)
{
    dyn_string dsAlertTexts;
    dyn_string dsAlertMatch;
    dyn_string dsAlertClasses;
    dyn_bool dbDiscreteNegation;
    dyn_string dsStateMatch;

    if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TransferJobs + "*", sSysAlarmDpe)) {
        dsAlertTexts = makeDynString(
                           "OK",
                           "Disabled transfer jobs detected",
                           "Transfer jobs not executed in time",
                           "Query Error");
        dsAlertMatch = makeDynString(
                           "*",
                           c_SysInt_LHCLoggingDB_TransferJobs_ErrorDisabled,
                           c_SysInt_LHCLoggingDB_TransferJobs_ErrorTimeout,
                           c_SysInt_LHCLoggingDB_TransferJobs_ErrorQuery);
        dsAlertClasses = makeDynString(
                             "",
                             "_unSystemIntegrityAlarm_lvl1.",
                             "_unSystemIntegrityAlarm_lvl1.",
                             "_unSystemIntegrityAlarm_lvl1.");
        dbDiscreteNegation = makeDynBool(false, false, false, false);
        dsStateMatch = makeDynString("", "", "", "");
    }
    else if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints + "*", sSysAlarmDpe)) {
        dsAlertTexts = makeDynString(
                           "OK",
                           "Transfer Error",
                           "Registration Error",
                           "Query Error");
        dsAlertMatch = makeDynString(
                           "*",
                           c_SysInt_LHCLoggingDB_TestDatapoints_ErrorTransfer,
                           c_SysInt_LHCLoggingDB_TestDatapoints_ErrorRegistration,
                           c_SysInt_LHCLoggingDB_TestDatapoints_ErrorQuery);
        dsAlertClasses = makeDynString(
                             "",
                             "_unSystemIntegrityAlarm_lvl1.",
                             "_unSystemIntegrityAlarm_lvl1.",
                             "_unSystemIntegrityAlarm_lvl1.");
        dbDiscreteNegation = makeDynBool(false, false, false, false);
        dsStateMatch = makeDynString("", "", "", "");
    }
    else if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_DataConsistency + "*", sSysAlarmDpe)) {
        dsAlertTexts = makeDynString(
                           "OK",
                           "Signals with inconsistent archive group detected",
                           "Signals with missing DPEs detected",
                           "Query Error");
        dsAlertMatch = makeDynString(
                           "*",
                           c_SysInt_LHCLoggingDB_DataConsistency_ErrorArchiveGroup,
                           c_SysInt_LHCLoggingDB_DataConsistency_ErrorMissingDpe,
                           c_SysInt_LHCLoggingDB_TestDatapoints_ErrorQuery);
        dsAlertClasses = makeDynString(
                             "",
                             "_unSystemIntegrityAlarm_lvl1.",
                             "_unSystemIntegrityAlarm_lvl1.",
                             "_unSystemIntegrityAlarm_lvl1.");
        dbDiscreteNegation = makeDynBool(false, false, false, false);
        dsStateMatch = makeDynString("", "", "", "");
    }
    else if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_FullConsistency + "*", sSysAlarmDpe)) {
        dsAlertTexts = makeDynString(
                           "OK",
                           "Inconsistent logging configuration",
                           "Error during full consistency check");
        dsAlertMatch = makeDynString(
                           "*",
                           c_SysInt_LHCLoggingDB_DataConsistency_ErrorIncosistent,
                           c_SysInt_LHCLoggingDB_FullConsistency_ErrorDuringCheck);
        dsAlertClasses = makeDynString(
                             "",
                             "_unSystemIntegrityAlarm_lvl1.",
                             "_unSystemIntegrityAlarm_lvl1.");
        dbDiscreteNegation = makeDynBool(false, false, false);
        dsStateMatch = makeDynString("", "", "");
    }
    else {
        DebugTN("Invalid alert dp name: " + sSysAlarmDpe);
        return;
    }

    // set discrete alarm config
    dyn_mixed alarmObject;
    fwAlertConfig_objectCreateDiscrete(
        alarmObject,
        dsAlertTexts,
        dsAlertMatch,
        dsAlertClasses,
        "",
        makeDynString(""),
        "",
        false,
        dbDiscreteNegation,
        "",
        dsStateMatch,
        dsExInfo,
        false,
        false,
        "",
        false); //exception info returned here
    if (dynlen(dsExInfo)) { return;}
    //set the alarm to the dpe
    fwAlertConfig_objectSet(sSysAlarmDpe, alarmObject, dsExInfo);
    if (dynlen(dsExInfo)) { return;}
    //activate it
    fwAlertConfig_activate(sSysAlarmDpe, dsExInfo);
    if (dynlen(dsExInfo)) { return;}
}

//------------------------------------------------------------------------------------------------------------------------

/**
* This function is used to handle the monitorig thread which does the checking
*
* @param command         input, string: "start" or "stop" the thread
*/
private void unSystemIntegrity_LHCLoggingDB_manageMonitoringThread(string command)
{
    switch (command) {
        case "start": {
            // start only if thread is not running yet
            if (g_SysInt_LHCLoggingDB_monitoringThreadId == 0) {
                g_SysInt_LHCLoggingDB_monitoringThreadId = startThread("unSystemIntegrity_LHCLoggingDB_monitoringThread");
            }
            break;
        }
        case "stop": {
            // stop thread only if necessary
            if (g_SysInt_LHCLoggingDB_monitoringThreadId != 0) {
                bool bStop = true;
                dyn_string dsSysIntAlarmDps = dpNames(c_unSystemAlarm_dpPattern + LHCLoggingDB_pattern + "*", c_unSystemAlarm_dpType);
                int iDsSysIntAlarmDpsLen = dynlen(dsSysIntAlarmDps);
                for (int i = 1; i <= iDsSysIntAlarmDpsLen; i++) {
                    bool bEnable;
                    dpGet(dsSysIntAlarmDps[i] + ".enabled", bEnable);
                    if (bEnable) {
                        bStop = false;
                    }
                }
                if (bStop) {
                    if (stopThread(g_SysInt_LHCLoggingDB_monitoringThreadId) < 0) {
                        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorStoppingThread);
                    } else {
                        g_SysInt_LHCLoggingDB_monitoringThreadId = 0;
                    }
                }
            }
            break;
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

/**
* This is the function that actually does the system integrity checking in a periodic cycle
*/
private void unSystemIntegrity_LHCLoggingDB_monitoringThread()
{
    unSystemIntegrity_LHCLoggingDB_registerDbSettingsChangedCB();

    while (1) {
        unSystemIntegrity_LHCLoggingDB_runChecks();
        delay(g_SysInt_LHCLoggingDB_loopDelay);
    }
}

//------------------------------------------------------------------------------------------------------------------------

private void unSystemIntegrity_LHCLoggingDB_registerDbSettingsChangedCB()
{
    // Reconnect to database if db settings change
    string sSettingsDpe;
    if (unLHCLoggingDBSettings_getSettingsDpeName(sSettingsDpe) == false) {
        DebugTN("Settings Datapoint not found");
        return;
    }
    dpConnect("unSystemIntegrity_LHCLoggingDB_dbSettingsChangedCB", false, sSettingsDpe);
    dyn_errClass err = getLastError();
    if (dynlen(err) > 0) {
        DebugTN("Error on dpConnect: ", err);
        return;
    }
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Reconnect to DB, if settings change
*/
private void unSystemIntegrity_LHCLoggingDB_dbSettingsChangedCB(string sDpe, string sValue)
{
    unSystemIntegrity_LHCLoggingDB_initDatabase();
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Run all enabled checks and report results to alarm data points.
*/
private void unSystemIntegrity_LHCLoggingDB_runChecks() synchronized(g_SysInt_LHCLoggingDB_mutex)
{
    string sAlarmDp;
    bool bAlarmEnabled;
    unsigned uResult;

    // 1. check status of transfer jobs
    sAlarmDp = c_unSystemAlarm_dpPattern + c_SysInt_LHCLoggingDB_IntegrityClass_TransferJobs;
    if (dpExists(sAlarmDp)) {
        dpGet(sAlarmDp + ".enabled", bAlarmEnabled);
        if (bAlarmEnabled) {
            uResult = unSystemIntegrity_LHCLoggingDB_checkTransferJobs(g_SysInt_LHCLoggingDB_mySystemDbSchema);
            dpSetWait(sAlarmDp + ".alarm", uResult);
        }
    }

    // 2. check history of test datapoints
    sAlarmDp = c_unSystemAlarm_dpPattern + c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints;
    if (dpExists(sAlarmDp)) {
        dpGet(sAlarmDp + ".enabled", bAlarmEnabled);
        if (bAlarmEnabled) {
            uResult = unSystemIntegrity_LHCLoggingDB_checkTestDatapoints();
            dpSetWait(sAlarmDp + ".alarm", uResult);
        }
    }

    // 3. check if all registered elements have an entry in ELEMENTS table
    sAlarmDp = c_unSystemAlarm_dpPattern + c_SysInt_LHCLoggingDB_IntegrityClass_DataConsistency;
    if (dpExists(sAlarmDp)) {
        dpGet(sAlarmDp + ".enabled", bAlarmEnabled);
        if (bAlarmEnabled) {
            uResult = unSystemIntegrity_LHCLoggingDB_checkDataConsistency(g_SysInt_LHCLoggingDB_mySystemDbSchema);
            dpSetWait(sAlarmDp + ".alarm", uResult);
        }
    }

    // 4. Full consistency check (only once a day)
    sAlarmDp = c_unSystemAlarm_dpPattern + c_SysInt_LHCLoggingDB_IntegrityClass_FullConsistency;
    if (dpExists(sAlarmDp)) {
        dpGet(sAlarmDp + ".enabled", bAlarmEnabled);
        if (bAlarmEnabled) {
            if(daySecond(getCurrentTime()) >= g_SysInt_LHCLoggingDB_fullConsistencyCheckDaySecond) {
                if(!g_SysInt_LHCLoggingDB_fullConsistencyCheckLocked) {
                    g_SysInt_LHCLoggingDB_fullConsistencyCheckLocked = true;
                    dyn_string dsExInfo;
                    dyn_string dsConfigurations = unLHCLoggingDBSettings_getConfigFileGeneratorRules(false, true);
                    unSystemIntegrity_LHCLoggingDB_doFullConsistencyCheck(dsConfigurations, true, true, false, dsExInfo);
                }
            } else {
                g_SysInt_LHCLoggingDB_fullConsistencyCheckLocked = false;
            }
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Check current status for all transfer jobs related to sDbSchema
*
* @param sDbSchema        input, database schema for which we want to check the transfer jobs
*
* @return uReturnCode     output, Statuscode
*/
private unsigned unSystemIntegrity_LHCLoggingDB_checkTransferJobs(string sDbSchema)
{
    dyn_string dsExInfo;

    // ensure a valid db connection
    unSystemIntegrity_LHCLoggingDB_initDatabase(false);

    unsigned uReturnCode = c_SysInt_LHCLoggingDB_TransferJobs_OK;

    time tDbSysTime = unLHCLoggingDBSql_getDbSysTime(dsExInfo);
    if (dynlen(dsExInfo) > 0) { return c_SysInt_LHCLoggingDB_TransferJobs_ErrorQuery; }

    dyn_dyn_mixed ddmJobStatus = unLHCLoggingDBSql_getTransferJobStatus(sDbSchema + ".%", dsExInfo);
    if (dynlen(dsExInfo) > 0) { return c_SysInt_LHCLoggingDB_TransferJobs_ErrorQuery; }

    int iDdmJobStatusLen = dynlen(ddmJobStatus);
    for (int i = 1; i <= iDdmJobStatusLen; i++) {
        if (ddmJobStatus[i][2] != "TRUE") {
            unSystemIntegrity_LHCLoggingDB_setReturnCode(uReturnCode, c_SysInt_LHCLoggingDB_TransferJobs_ErrorDisabled);
            continue;
        }

        time tLastJobStart = ddmJobStatus[i][4];
        if (tDbSysTime - tLastJobStart > g_SysInt_LHCLoggingDB_maxAgeTransferJobs) {
            unSystemIntegrity_LHCLoggingDB_setReturnCode(uReturnCode, c_SysInt_LHCLoggingDB_TransferJobs_ErrorTimeout);
            continue;
        }
    }

    return uReturnCode;
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Check current status for all test datapoints
*
* @return uReturnCode     output, Statuscode
*/
private unsigned unSystemIntegrity_LHCLoggingDB_checkTestDatapoints()
{
    // ensure a valid db connection
    unSystemIntegrity_LHCLoggingDB_initDatabase(false);

    // Get variable names of test data points
    dyn_string dsVariableName = unSystemIntegrity_LHCLoggingDB_getTestDpVarNames();

    dyn_string dsExInfo;
    unsigned uReturnCode = c_SysInt_LHCLoggingDB_TestDatapoints_OK;

    time tDbSysTime = unLHCLoggingDBSql_getDbSysTime(dsExInfo);
    if (dynlen(dsExInfo) > 0) { return c_SysInt_LHCLoggingDB_TestDatapoints_ErrorQuery; }

    for (int i=1; i<=dynlen(dsVariableName); i++) {

        // Check transfer history, if there are no values we are initialising
        dyn_dyn_mixed ddmTransferHistory = unLHCLoggingDBSql_getTransferHistoryForTestDpe(dsVariableName[i], dsExInfo);
        if (dynlen(dsExInfo) > 0) { return c_SysInt_LHCLoggingDB_TestDatapoints_ErrorQuery; }

        // Check if element is registered, if no, set registration error
        if (dynlen(ddmTransferHistory) == 0) {
            unSystemIntegrity_LHCLoggingDB_setReturnCode(uReturnCode, c_SysInt_LHCLoggingDB_TestDatapoints_ErrorRegistration);
            continue;
        }

        // Does result have an LAST_LOGGED_UTC timestamp?
        if (ddmTransferHistory[1][2] == "") {
            unSystemIntegrity_LHCLoggingDB_setReturnCode(uReturnCode, c_SysInt_LHCLoggingDB_TestDatapoints_ErrorTransfer);
            continue;
        }

        // Is LAST_LOGGED_UTC more recent than maxAgeTransferHistory?
        time tLastLogged = ddmTransferHistory[1][2];
        if (tDbSysTime - tLastLogged > g_SysInt_LHCLoggingDB_maxAgeTransferHistory) {
            unSystemIntegrity_LHCLoggingDB_setReturnCode(uReturnCode, c_SysInt_LHCLoggingDB_TestDatapoints_ErrorTransfer);
            continue;
        }

        unSystemIntegrity_LHCLoggingDB_setReturnCode(uReturnCode, c_SysInt_LHCLoggingDB_TestDatapoints_OK);
    }

    return uReturnCode;
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Check data consistecy between registered variables in database (table: META_VARIABLES)
* and datapoints with an archive config in the PVSS project
*
* @param sDbSchema       input, DB schema to check
*
* @return uReturnCode    output, Statuscode
*/
private unsigned unSystemIntegrity_LHCLoggingDB_checkDataConsistency(string sDbSchema)
{
    dyn_string dsExInfo;

    // ensure a valid db connection
    unSystemIntegrity_LHCLoggingDB_initDatabase(false);

    dyn_dyn_mixed ddmData = unLHCLoggingDBSql_getInconsistentData(sDbSchema, dsExInfo);

    unsigned numSigWithIncArchiveGroup = unLHCLoggingDBSql_getNumSignalsHavingInconsistentArchiveGroup(sDbSchema, dsExInfo);

    if (dynlen(dsExInfo) > 0) { return c_SysInt_LHCLoggingDB_DataConsistency_ErrorQuery; }
    if (dynlen(ddmData) > 0) { return c_SysInt_LHCLoggingDB_DataConsistency_ErrorMissingDpe; }
    if(numSigWithIncArchiveGroup > 0) { return c_SysInt_LHCLoggingDB_DataConsistency_ErrorArchiveGroup; }
    return c_SysInt_LHCLoggingDB_DataConsistency_OK;
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Checks wheter the new return code has higher priority as the current one,
* if yes, the current one will be set to the new one.
*
* @param current     input/output, return code
* @param new         input, new return code
*/
private void unSystemIntegrity_LHCLoggingDB_setReturnCode(unsigned &current, unsigned uNew) {
    if(uNew > current) {
        current = uNew;
    }
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Write special error to message log
*
* @param severity     input, severity of the message
* @param code         input, error code to write
* @param sDetails     input, message details
*/
private void unSystemIntegrity_LHCLoggingDB_throwError(string severity = "ERROR", int code = 1, string sDetails = "")
{
    int prio = PRIO_WARNING; // it seems that throwing a fatal error makes Pmon stop monitoring this manager????
    int type = ERR_CONTROL;

    switch (strtoupper(severity)) {
        case "INFO":
            prio = PRIO_INFO;
            break;
        case "WARNING":
            prio = PRIO_WARNING;
            break;
        case "ERROR":
            prio = PRIO_SEVERE;
            break;
    }

    errClass err = makeError("unSystemIntegrity_LHCLoggingDB", prio, type, code, sDetails);
    throwError(err);
}

//------------------------------------------------------------------------------------------------------------------------

/**
*  Checks consistency between import files and configuration stored in DB
*/
private void unSystemIntegrity_LHCLoggingDB_doFullConsistencyCheck(
        dyn_string dsConfigurations,     // input, list of configurations to check
        bool bGenerateTempImportFiles,   // input, also regenerate the config files
        bool bSetAlarmDP,                // input, set the system integrity alert
        bool bWriteResultToDP,           // input, write result of the check to a datapoint
        dyn_string &dsExInfo)            // output, exception information
{
    // ensure a valid db connection
    unSystemIntegrity_LHCLoggingDB_initDatabase(false);

    unsigned uAlarmValue = c_SysInt_LHCLoggingDB_FullConsistency_ErrorDuringCheck;

    DebugFTN(1, __FUNCTION__+":"+__LINE__, "check started");
    DebugFTN(1, "Configurations: " + dynStringToString(dsConfigurations),
                "GenerateTempImportFiles: " + bGenerateTempImportFiles,
                "SetAlarmDP: " + bSetAlarmDP,
                "WriteResultToDP: " + bWriteResultToDP);

    try {
        dyn_string dsImportFiles;
        if(bGenerateTempImportFiles) {
            dsImportFiles = unLHCLoggingDBParser_generateImportFiles(dsConfigurations, true);  // this throws an exception if user function is not found
        } else {
            dsImportFiles = unLHCLoggingDBParser_getImportFileNames(dsConfigurations);
        }

        mapping mDbConfig = unSystemIntegrity_LHCLoggingDB_getDbConfig(dsExInfo);
        if (dynlen(dsExInfo) == 0) {
            mapping mFileConfig = unSystemIntegrity_LHCLoggingDB_getFileConfig(dsImportFiles, dsExInfo);
            if (dynlen(dsExInfo) == 0) {
                mapping mResult = unSystemIntegrity_LHCLoggingDB_compareConfig(mDbConfig, mFileConfig);
                if(bWriteResultToDP) {
                    unSystemIntegrity_LHCLoggingDB_writeResultToDP("unLHCLoggingDB_ConsistencyCheck", mResult);
                }
                uAlarmValue = (mResult["Status"] ? c_SysInt_LHCLoggingDB_FullConsistency_OK : c_SysInt_LHCLoggingDB_DataConsistency_ErrorIncosistent);
            } else {
                DebugFTN(1, __FUNCTION__+":"+__LINE__, dsExInfo);
            }
        } else {
            DebugFTN(1, __FUNCTION__+":"+__LINE__, dsExInfo);
        }
    } catch {
        DebugTN(getLastException());
        fwException_raise(dsExInfo, "ERROR", "Exception occured", "10");
    }

    if (bSetAlarmDP) {
        unSystemIntegrity_LHCLoggingDB_setAlarmDP(c_SysInt_LHCLoggingDB_IntegrityClass_FullConsistency, uAlarmValue);
    }

    DebugFTN(1, __FUNCTION__+":"+__LINE__, "check finished");
}

//------------------------------------------------------------------------------------------------------------------------


private void unSystemIntegrity_LHCLoggingDB_setAlarmDP(string sIntegrityClass, unsigned uAlarmValue)
{
    const string sAlarmDp = c_unSystemAlarm_dpPattern + sIntegrityClass;

    if (dpExists(sAlarmDp)) {
        bool bAlarmEnabled;
        dpGet(sAlarmDp + ".enabled", bAlarmEnabled);
        if (bAlarmEnabled) {
            dpSetWait(sAlarmDp + ".alarm", uAlarmValue);
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

/**
* Get logging configuration from database
*/
private mapping unSystemIntegrity_LHCLoggingDB_getDbConfig(dyn_string &dsExInfo) synchronized(g_SysInt_LHCLoggingDB_mutex)
{
    mapping mResult;

    mapping mElements;
    mElements["NAME"] = makeDynString();

    mapping mMetaVars;
    mMetaVars["NAME"] = makeDynString();
    mMetaVars["ELEMENT"] = makeDynString();
    mMetaVars["DESCRIPTION"] = makeDynString();
    mMetaVars["HIERARCHY"] = makeDynString();

    string sDbSchema = unLHCLoggingDBSettings_getMySystemDbSchema();

    // Get elements from MDB
    dyn_dyn_mixed ddmElements = unLHCLoggingDBSql_getElements(sDbSchema, dsExInfo);
    if (dynlen(dsExInfo) > 0) {
        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorFullConsistencyCheck, "Elements");
        return mResult;
    }
    if (dynlen(ddmElements[1]) > 0) {
        mElements["NAME"] = ddmElements[1];
    }

    // Get registered variables
    dyn_dyn_mixed ddmMetaVars = unLHCLoggingDBSql_getLogVariables(sDbSchema, dsExInfo);
    if (dynlen(dsExInfo) > 0) {
        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorFullConsistencyCheck, "Meta Variables");
        return mResult;
    }
    if (dynlen(ddmMetaVars) > 0) {
        mMetaVars["NAME"] = ddmMetaVars[1];
        mMetaVars["ELEMENT"] = ddmMetaVars[2];
        mMetaVars["DESCRIPTION"] = ddmMetaVars[3];
        mMetaVars["HIERARCHY"] = ddmMetaVars[4];
        for (unsigned iHierarchy = dynlen(mMetaVars["HIERARCHY"]); iHierarchy>0; iHierarchy--) {
          strreplace(mMetaVars["HIERARCHY"][iHierarchy], "->", "/"); // Ensure db config looks like file config
        }
    }

    mResult["ELEMENTS"] = mElements;
    mResult["META_VARS"] = mMetaVars;
    return mResult;
}

/**
* Get logging configuration from config files
*/
private mapping unSystemIntegrity_LHCLoggingDB_getFileConfig(const dyn_string &dsImportFiles, dyn_string &dsExInfo)
{
    mapping mResult;
    mResult["NAME"] = makeDynString();
    mResult["ELEMENT"] = makeDynString();
    mResult["DESCRIPTION"] = makeDynString();
    mResult["HIERARCHY"] = makeDynString();
    mResult["DATA_CATEG"] = makeDynString();
    mResult["FILE"] = makeDynString();

    // Read configuration data from files
    for (int i = 1; i <= dynlen(dsImportFiles); i++) {
        dyn_string dsVarName, dsDpe, dsHierarchy, dsDescription, dsDataCategory, dsConfigFile;
        bool success = unLHCLoggingDBParser_getConfigFileData(getPath(DATA_REL_PATH)+dsImportFiles[i], dsVarName, dsDpe, dsHierarchy, dsDescription, dsDataCategory);
        if(!success) {
            fwException_raise(dsExInfo, "ERROR", "Cannot read config file: " + dsImportFiles[i], "10");
            return mResult;
        }

        int len = dynlen(dsVarName);
        for (int j = 1; j <= len; j++) {
            dsConfigFile[j] = dsImportFiles[i];
            // Change names to upper case cause they are stored uppercase letters in db and we have to compare the strings
            dsVarName[j] = strtoupper(dsVarName[j]);
            dsHierarchy[j] = strtoupper(dsHierarchy[j]);
        }

        dynAppend(mResult["NAME"], dsVarName);
        dynAppend(mResult["ELEMENT"], dsDpe);
        dynAppend(mResult["DESCRIPTION"], dsDescription);
        dynAppend(mResult["HIERARCHY"], dsHierarchy);
        dynAppend(mResult["DATA_CATEG"], dsDataCategory);
        dynAppend(mResult["FILE"], dsConfigFile);
    }

    // Add test datapoints to the list
    dyn_string dsTestDp = unSystemIntegrity_LHCLoggingDB_getTestDps();
    int len = dynlen(dsTestDp);
    for (int i = 1; i <= len; i++) {
        string sDataCategory = unSystemIntegrity_LHCLoggingDB_getDataCategory(dsTestDp[i]);
        strreplace(sDataCategory, "/", "_");
        string sDpeName = dsTestDp[i] + c_SysInt_RDB_TestDpe_Value;
        string sVarName = unSystemIntegrity_LHCLoggingDB_getVarName(dsTestDp[i]);
        string sDescription = dpGetDescription(sDpeName);

        dynAppend(mResult["NAME"], sVarName);
        dynAppend(mResult["ELEMENT"], sDpeName);
        dynAppend(mResult["DESCRIPTION"], sDescription);
        dynAppend(mResult["HIERARCHY"], "");
        dynAppend(mResult["DATA_CATEG"], sDataCategory);
        dynAppend(mResult["FILE"], "TestDPs autom. created");
    }

    return mResult;
}

/**
* Compare logging configuration
*/
private mapping unSystemIntegrity_LHCLoggingDB_compareConfig(const mapping &mDbConfig, const mapping &mFileConfig)
{
    mapping mResult;
    mResult["Status"] = true;
    mResult["File2Db"] = unSystemIntegrity_LHCLoggingDB_compareFileWithDb(mDbConfig, mFileConfig);
    mResult["Db2File"] = unSystemIntegrity_LHCLoggingDB_compareDbWithFile(mDbConfig, mFileConfig);
    mResult["Status"] = ((mResult["File2Db"]["Errors"] + mResult["Db2File"]["Errors"]) == 0);
    return mResult;
}

/**
* Compare logging config from file with logging config from db
*/
private mapping unSystemIntegrity_LHCLoggingDB_compareFileWithDb(const mapping &mDbConfig, const mapping &mFileConfig)
{
    mapping mResult;
    mResult["VarName"] = makeDynString();
    mResult["Exist"] = makeDynString();
    mResult["Match"] = makeDynString();
    mResult["Info"] = makeDynString();
    mResult["ConfigFile"] = makeDynString();
    mResult["Errors"] = 0;

    // Compare: FILE --> DB
    // =================  //
    int len = dynlen(mFileConfig["NAME"]);
    for (int i = 1; i <= len; i++) {
        mResult["VarName"][i] = mFileConfig["NAME"][i];
        mResult["Exist"][i] = "NO";
        mResult["Match"][i] = "NO";
        mResult["Info"][i] = "";
        mResult["ConfigFile"][i] = mFileConfig["FILE"][i];

        // Check if Variable_Name exists in database
        int dbIdx = dynContains(mDbConfig["META_VARS"]["NAME"], mFileConfig["NAME"][i]);
        if (dbIdx <= 0) {
            mResult["Info"][i] = "Signal not registered for logging.";
            mResult["Errors"] += 1;
            continue;
        }
        mResult["Exist"][i] = "YES";

        // Check if Dpe exists on system
        if (mFileConfig["ELEMENT"][i] == "") {
            mResult["Info"][i] = "DP does not exist on system.";
            mResult["Errors"] += 1;
            continue;
        }

        // Check if Dpe exists in elements table
        if (dynContains(mDbConfig["ELEMENTS"]["NAME"], mFileConfig["ELEMENT"][i]) <= 0) {
            mResult["Info"][i] = "DP does not exist in RDB Archive.";
            mResult["Errors"] += 1;
            continue;
        }

        // Check if Dpe from config file matches Element_Name from db
        if (mFileConfig["ELEMENT"][i] != mDbConfig["META_VARS"]["ELEMENT"][dbIdx]) {
            mResult["Info"][i] = "DP does not match with registered DP.";
            mResult["Errors"] += 1;
            continue;
        }

        //Check if hierarchy from config file matches hierarchy in db
        if (mFileConfig["HIERARCHY"][i] != strtoupper((string)mDbConfig["META_VARS"]["HIERARCHY"][dbIdx])) {
            mResult["Info"][i] = "Hierarchy does not match. " +
                                            "Config file: '" + mFileConfig["HIERARCHY"][i] + "' " +
                                            "Database: '" + mDbConfig["META_VARS"]["HIERARCHY"][dbIdx] + "' ";
            mResult["Errors"] += 1;
            continue;
        }

        mResult["Match"][i] = "YES";
        mResult["Info"][i] = "";
    }

    return mResult;
}

/**
* Compare logging config from db with logging config from file
*/
private mapping unSystemIntegrity_LHCLoggingDB_compareDbWithFile(const mapping &mDbConfig, const mapping &mFileConfig)
{
    mapping mResult;
    mResult["VarName"] = makeDynString();
    mResult["Exist"] = makeDynString();
    mResult["Match"] = makeDynString();
    mResult["Info"] = makeDynString();
    mResult["Errors"] = 0;

    // Compare: DB --> FILE
    // =================  //
    int len = dynlen(mDbConfig["META_VARS"]["NAME"]);
    for (int i = 1; i <= len; i++) {
        mResult["VarName"][i] = mDbConfig["META_VARS"]["NAME"][i];
        mResult["Exist"][i] = "NO";
        mResult["Match"][i] = "NO";
        mResult["Info"][i] = "";

        // Check if Variable_Name from database is found in config file
        int fileIdx = dynContains(mFileConfig["NAME"], mDbConfig["META_VARS"]["NAME"][i]);
        if (fileIdx <= 0) {
            mResult["Info"][i] = "Signal not found in config file.";
            mResult["Errors"] += 1;
            continue;
        }
        mResult["Exist"][i] = "YES";

        // Check if Element_Name for this Variable_Name exists in database
        if (mDbConfig["META_VARS"]["ELEMENT"][i] == "") {
            mResult["Info"][i] = "Signal has no associated DP in RDB Archive.";
            mResult["Errors"] += 1;
            continue;
        }

        // Check if Element_Name exists on this system
        if (!dpExists(mDbConfig["META_VARS"]["ELEMENT"][i])) {
            mResult["Info"][i] = "DP does not exist on system.";
            mResult["Errors"] += 1;
            continue;
        }

        // Check if Dpe from DB matches dp from config file
        if (mDbConfig["META_VARS"]["ELEMENT"][i] != mFileConfig["ELEMENT"][fileIdx]) {
            mResult["Info"][i] = "DP does not match with config file.";
            mResult["Errors"] += 1;
            continue;
        }

        //Check if hierarchy from config file matches hierarchy in db
        if (strtoupper((string)mDbConfig["META_VARS"]["HIERARCHY"][i]) != mFileConfig["HIERARCHY"][fileIdx]) {
            mResult["Info"][i] = "Hierarchy does not match. " +
                                            "Config file: '" + mFileConfig["HIERARCHY"][fileIdx] + "' " +
                                            "Database: '" + mDbConfig["META_VARS"]["HIERARCHY"][i] + "' ";
            mResult["Errors"] += 1;
            continue;
        }

        mResult["Match"][i] = "YES";
        mResult["Info"][i] = "";
    }

    return mResult;
}

/**
* output result to datapoint
*/
private void unSystemIntegrity_LHCLoggingDB_writeResultToDP(const string sResultDP, const mapping &mResult)
{
    dpSetWait(sResultDP + ".File2Db.VarName", mResult["File2Db"]["VarName"],
              sResultDP + ".File2Db.Exist", mResult["File2Db"]["Exist"],
              sResultDP + ".File2Db.Match", mResult["File2Db"]["Match"],
              sResultDP + ".File2Db.Info", mResult["File2Db"]["Info"],
              sResultDP + ".File2Db.ConfigFile", mResult["File2Db"]["ConfigFile"],
              sResultDP + ".File2Db.Errors", mResult["File2Db"]["Errors"],
              sResultDP + ".Db2File.VarName", mResult["Db2File"]["VarName"],
              sResultDP + ".Db2File.Exist", mResult["Db2File"]["Exist"],
              sResultDP + ".Db2File.Match", mResult["Db2File"]["Match"],
              sResultDP + ".Db2File.Info", mResult["Db2File"]["Info"],
              sResultDP + ".Db2File.Errors", mResult["Db2File"]["Errors"],
              sResultDP + ".Status", mResult["Status"] );
}
